@@ -14,7 +14,6 @@ from .errors import NnenumError, NnenumTranslatorError
 
 
 class Nnenum(Verifier):
-    EXE = "nnenum.py"
     reduction = partial(IOPolytopeReduction, HalfspacePolytope, HalfspacePolytope)
     translator_error = NnenumTranslatorError
     verifier_error = NnenumError
@@ -49,7 +48,7 @@ class Nnenum(Verifier):
         with tempfile.NamedTemporaryFile(
             mode="w+", suffix=".onnx", delete=False
         ) as onnx_model_file:
-            prop.op_graph.export_onnx(onnx_model_file.name)
+            prop.op_graph.simplify().export_onnx(onnx_model_file.name)
 
         lb, ub = prop.input_constraint.as_bounds()
         A_in, b_in = prop.input_constraint.as_matrix_inequality()
@@ -65,7 +64,7 @@ class Nnenum(Verifier):
         ) as output_file:
             self._tmp_output_file = output_file
         args = (
-            "nnenum.py",
+            "nnenum",
             onnx_model_file.name,
             constraint_file.name,
             "-o",
@@ -83,6 +82,8 @@ class Nnenum(Verifier):
         result_str, cex = np.load(self._tmp_output_file.name, allow_pickle=True)
         if result_str == "safe":
             return UNSAT, None
-        elif result_str == "unsafe":
+        elif result_str.startswith("unsafe"):
             return SAT, cex
+        elif result_str == "error":
+            raise self.verifier_error("result:error")
         raise self.translator_error(f"Unknown verification result: {result_str}")
